#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.symbols


def test_parse_symbol():
    assert octobot_commons.symbols.parse_symbol("BTC/USDT") == octobot_commons.symbols.Symbol("BTC/USDT")


def test_merge_symbol():
    assert octobot_commons.symbols.merge_symbol("BTC/USDT") == "BTCUSDT"
    assert octobot_commons.symbols.merge_symbol("BTC/USDT:USDT") == "BTCUSDT_USDT"


def test_merge_currencies():
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT") == "BTC/USDT"
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT", "BTC") == "BTC/USDT:BTC"
    assert octobot_commons.symbols.merge_currencies("BTC", "USDT", "XXX", "g", "d") == "BTCgUSDTdXXX"


def test_convert_symbol():
    assert octobot_commons.symbols.convert_symbol("BTC-USDT", symbol_separator="-") == "BTC/USDT"



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import jsonschema


def validate(config, schema_file) -> None:
    """
    Validate a config file, raise upon validation error
    :param config: the config
    :param schema_file: the config schema
    :return: None
    """
    with open(schema_file) as json_schema:
        loaded_schema = json.load(json_schema)
    jsonschema.validate(instance=config, schema=loaded_schema)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.configuration import configuration
from octobot_commons.configuration import config_file_manager
from octobot_commons.configuration import config_operations
from octobot_commons.configuration import fields_utils
from octobot_commons.configuration import user_inputs


from octobot_commons.configuration.configuration import (
    Configuration,
)
from octobot_commons.configuration.config_file_manager import (
    get_user_config,
    load,
    dump,
    check_config,
    jsonify_config,
    handle_encrypted_value,
    prepare_restore_file,
    remove_restore_file,
    restore,
    dump_formatted_json,
)
from octobot_commons.configuration.config_operations import (
    filter_to_update_data,
    parse_and_update,
    merge_dictionaries_by_appending_keys,
    clear_dictionaries_by_keys,
)
from octobot_commons.configuration.fields_utils import (
    has_invalid_default_config_value,
    encrypt,
    decrypt,
    decrypt_element_if_possible,
    get_password_hash,
)
from octobot_commons.configuration.user_inputs import (
    UserInput,
    UserInputFactory,
    sanitize_user_input_name,
    save_user_input,
    get_user_input_tentacle_type,
    get_user_inputs,
    clear_user_inputs,
)


__all__ = [
    "Configuration",
    "get_user_config",
    "load",
    "dump",
    "check_config",
    "jsonify_config",
    "handle_encrypted_value",
    "prepare_restore_file",
    "remove_restore_file",
    "restore",
    "dump_formatted_json",
    "filter_to_update_data",
    "parse_and_update",
    "merge_dictionaries_by_appending_keys",
    "clear_dictionaries_by_keys",
    "has_invalid_default_config_value",
    "encrypt",
    "decrypt",
    "decrypt_element_if_possible",
    "get_password_hash",
    "UserInput",
    "UserInputFactory",
    "sanitize_user_input_name",
    "save_user_input",
    "get_user_input_tentacle_type",
    "get_user_inputs",
    "clear_user_inputs",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os
import zipfile
import shutil
import pathlib
import uuid
import time
import octobot_commons.constants as constants
import octobot_commons.logging as bot_logging

# avoid cyclic import
from octobot_commons.profiles.profile import Profile


def export_profile(profile, export_path: str) -> str:
    """
    Exports the given profile into export_path, appends ".zip" as a file extension
    :param profile: profile to export
    :param export_path: export path ending with filename
    :return: the exported profile path including file extension
    """
    temp_path = f"{export_path}{int(time.time() * 1000)}"
    # remove any existing file to prevent any side effect
    if os.path.exists(temp_path):
        raise OSError(f"Can't export profile, the {temp_path} folder exists")
    export_path_with_ext = f"{export_path}.{constants.PROFILE_EXPORT_FORMAT}"
    if os.path.isfile(export_path_with_ext):
        os.remove(export_path_with_ext)
    # copy profile into a temp dir to edit it
    shutil.copytree(profile.path, temp_path)
    try:
        _filter_profile_export(temp_path)
        # export the edited profile
        shutil.make_archive(export_path, constants.PROFILE_EXPORT_FORMAT, temp_path)
    finally:
        shutil.rmtree(temp_path)
    return export_path_with_ext


def import_profile(
    import_path: str,
    name: str = None,
    bot_install_path: str = ".",
    replace_if_exists: bool = False,
) -> Profile:
    """
    Imports the given profile export archive into the user's profile directory with the "imported_" prefix
    :param import_path: path to the profile zipped archive
    :param name: name of the profile folder
    :param bot_install_path: path to the octobot installation
    :param replace_if_exists: when True erase the profile with the same name if it exists
    :return: None
    """
    logger = bot_logging.get_logger("ProfileSharing")
    profile_name = name or (
        f"{constants.IMPORTED_PROFILE_PREFIX}_{os.path.split(import_path)[-1]}"
    )
    profile_name = profile_name.split(f".{constants.PROFILE_EXPORT_FORMAT}")[0]
    target_import_path, replaced = _get_target_import_path(
        bot_install_path, profile_name, replace_if_exists
    )
    action = "Creat"
    if replaced:
        action = "Updat"
    logger.info(f"{action}ing {profile_name} profile.")
    _import_profile_files(import_path, target_import_path)
    profile = Profile(target_import_path).read_config()
    profile.imported = True
    _ensure_unique_profile_id(profile)
    logger.info(f"{action}ed {profile.name} ({profile_name}) profile.")
    return profile


def _filter_profile_export(profile_path: str):
    profile_file = os.path.join(profile_path, constants.PROFILE_CONFIG_FILE)
    if os.path.isfile(profile_file):
        with open(profile_file) as open_file:
            parsed_profile = json.load(open_file)
        _filter_disabled(parsed_profile, constants.CONFIG_EXCHANGES)
        with open(profile_file, "w") as open_file:
            json.dump(parsed_profile, open_file, indent=4, sort_keys=True)


def _filter_disabled(profile_config: dict, element):
    filtered_exchanges = {
        exchange: details
        for exchange, details in profile_config[constants.PROFILE_CONFIG][
            element
        ].items()
        if details.get(constants.CONFIG_ENABLED_OPTION, True)
    }
    profile_config[constants.PROFILE_CONFIG][element] = filtered_exchanges


def _get_target_import_path(
    bot_install_path: str, profile_name: str, replace_if_exists: bool
) -> (str, bool):
    """
    Get the target profile folder path
    :param bot_install_path: path to the octobot installation
    :param profile_name: name of the profile folder
    :param replace_if_exists: when True erase the profile with the same name if it exists
    :return: (the final target import path, True if the profile is replaced)
    """
    target_import_path = os.path.join(
        bot_install_path, constants.USER_PROFILES_FOLDER, profile_name
    )
    if replace_if_exists:
        try:
            replaced = True
            shutil.rmtree(target_import_path)
        except FileNotFoundError:
            replaced = False
        return target_import_path, replaced
    return _get_unique_profile_folder(target_import_path), False


def _import_profile_files(profile_path: str, target_profile_path: str) -> None:
    """
    Copy or extract profile files to destination
    :param profile_path: the current profile path
    :param target_profile_path: the target profile path
    :return: None
    """
    if zipfile.is_zipfile(profile_path):
        with zipfile.ZipFile(profile_path) as zipped:
            zipped.extractall(target_profile_path)
    else:
        shutil.copytree(profile_path, target_profile_path)


def _get_unique_profile_folder(target_import_path: str) -> str:
    """
    Creates an unique profile folder name
    :param target_import_path: the expected target profile folder name
    :return: the unique profile folder name
    """
    iteration = 1
    candidate = target_import_path
    while os.path.exists(candidate) and iteration < 100:
        iteration += 1
        candidate = f"{target_import_path}_{iteration}"
    return candidate


def _ensure_unique_profile_id(profile) -> None:
    """
    Ensure that no other installed profile has the same id
    :param profile: the installed profile
    :return: None
    """
    ids = Profile.get_all_profiles_ids(
        pathlib.Path(profile.path).parent, ignore=profile.path
    )
    iteration = 1
    while profile.profile_id in ids and iteration < 100:
        profile.profile_id = str(uuid.uuid4())
        iteration += 1
    profile.save()



# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef bint default_parent_inspection(object element, object parent)
cpdef bint default_parents_inspection(object element, object parent)
cpdef bint evaluator_parent_inspection(object element, object parent)
cpdef bint trading_mode_parent_inspection(object element, object parent)

cpdef object get_class_from_parent_subclasses(str class_string, object parent)
cpdef object get_deep_class_from_parent_subclasses(str class_string, object parent)

cpdef bint is_abstract_using_inspection_and_class_naming(object clazz)

cpdef list get_all_classes_from_parent(object parent_class)
cpdef object get_single_deepest_child_class(object clazz)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import octobot_commons.enums as enums

# time
MSECONDS_TO_SECONDS = 1000
MINUTE_TO_SECONDS = 60
MSECONDS_TO_MINUTE = MSECONDS_TO_SECONDS * MINUTE_TO_SECONDS
HOURS_TO_SECONDS = MINUTE_TO_SECONDS * 60
HOURS_TO_MSECONDS = MSECONDS_TO_SECONDS * MINUTE_TO_SECONDS * MINUTE_TO_SECONDS
DAYS_TO_SECONDS = HOURS_TO_SECONDS * 24

# Strings
CONFIG_WILDCARD = "*"
CONFIG_SYMBOLS_WILDCARD = ["*"]
PORTFOLIO_AVAILABLE = "available"
MARGIN_PORTFOLIO = "margin"
PORTFOLIO_TOTAL = "total"

# config
CONFIG_ENABLED_OPTION = "enabled"
CONFIG_DEBUG_OPTION = "DEV-MODE"
CONFIG_TIME_FRAME = "time_frame"
USER_FOLDER = "user"
CONFIG_FOLDER = "config"
CONFIG_FILE = "config.json"
TEMP_RESTORE_CONFIG_FILE = "temp_config.json"
DEFAULT_CONFIG_FILE = "default_config.json"
DEFAULT_CONFIG_FILE_PATH = f"{CONFIG_FOLDER}/{DEFAULT_CONFIG_FILE}"
SCHEMA = "schema"
CONFIG_FILE_EXT = ".json"
CONFIG_FILE_SCHEMA = f"{CONFIG_FOLDER}/config_{SCHEMA}.json"
CONFIG_REFRESH_RATE = "refresh_rate_seconds"
CONFIG_SAVED_HISTORICAL_TIMEFRAMES = "saved_historical_timeframes"
CONFIG_OPTIMIZER_ID = "optimizer_id"
CONFIG_BACKTESTING_ID = "backtesting_id"
CONFIG_CURRENT_LIVE_ID = "current-live-id"
DEFAULT_CURRENT_LIVE_ID = 1

# profiles
PROFILES_FOLDER = "profiles"
USER_PROFILES_FOLDER = f"{USER_FOLDER}/{PROFILES_FOLDER}"
PROFILE_CONFIG_FILE = "profile.json"
CONFIG_PROFILE = "profile"
DEFAULT_PROFILE = "default"
DEFAULT_PROFILE_FILE = f"{CONFIG_PROFILE}.json"
CONFIG_NAME = "name"
CONFIG_DESCRIPTION = "description"
CONFIG_AVATAR = "avatar"
CONFIG_READ_ONLY = "read_only"
CONFIG_IMPORTED = "imported"
PROFILE_CONFIG = "config"
CONFIG_ID = "id"
PROFILE_FILE_SCHEMA = f"{CONFIG_FOLDER}/profile_{SCHEMA}.json"
PROFILE_EXPORT_FORMAT = "zip"
IMPORTED_PROFILE_PREFIX = "imported"
USE_CURRENT_PROFILE = "use_current_profile"

# Config currencies
CONFIG_CRYPTO_CURRENCIES = "crypto-currencies"
CONFIG_CRYPTO_CURRENCY = "crypto-currency"
CONFIG_CRYPTO_PAIRS = "pairs"
CONFIG_CRYPTO_QUOTE = "quote"
CONFIG_CRYPTO_ADD = "add"

# Exchange
CONFIG_EXCHANGES = "exchanges"
CONFIG_EXCHANGE_KEY = "api-key"
CONFIG_EXCHANGE_SECRET = "api-secret"
CONFIG_EXCHANGE_PASSWORD = "api-password"
CONFIG_EXCHANGE_TYPE = "exchange-type"
CONFIG_CONTRACT_TYPE = "contract-type"
CONFIG_EXCHANGE_SANDBOXED = "sandboxed"
CONFIG_EXCHANGE_FUTURE = "future"
CONFIG_EXCHANGE_MARGIN = "margin"
CONFIG_EXCHANGE_SPOT = "spot"
CONFIG_EXCHANGE_REST_ONLY = "rest_only"
CONFIG_EXCHANGE_WEB_SOCKET = "web-socket"
CONFIG_EXCHANGE_SUB_ACCOUNT = "sub_account"
CONFIG_EXCHANGE_ENCRYPTED_VALUES = [
    CONFIG_EXCHANGE_KEY,
    CONFIG_EXCHANGE_SECRET,
    CONFIG_EXCHANGE_PASSWORD,
]

# Trader
CONFIG_TRADING = "trading"
CONFIG_TRADER = "trader"
CONFIG_LOAD_TRADE_HISTORY = "load-trade-history"
CONFIG_TRADER_RISK = "risk"
CONFIG_TRADER_ALLOW_ARTIFICIAL_ORDERS = "allow-artificial-orders"
CONFIG_TRADER_RISK_MIN = 0.05
CONFIG_TRADER_RISK_MAX = 1
CONFIG_TRADER_REFERENCE_MARKET = "reference-market"

# Simulator
CONFIG_SIMULATOR = "trader-simulator"
CONFIG_STARTING_PORTFOLIO = "starting-portfolio"
SIMULATOR_CURRENT_PORTFOLIO = "simulator_current_portfolio"
CONFIG_SIMULATOR_FEES = "fees"
CONFIG_SIMULATOR_FEES_MAKER = "maker"
CONFIG_SIMULATOR_FEES_TAKER = "taker"
CONFIG_SIMULATOR_FEES_WITHDRAW = "withdraw"

# Optimization campaigns
DEFAULT_CAMPAIGN = "default_campaign"

# Optimizer
OPTIMIZER_RUNS_FOLDER = "optimizer"

# OS
PLATFORM_DATA_SEPARATOR = ":"
CLOCK_REFRESH_HOURS_INTERVAL = int(os.getenv("CLOCK_REFRESH_HOURS_INTERVAL", "4"))

# Evaluators
MIN_EVAL_TIME_FRAME = enums.TimeFrames.ONE_MINUTE
INIT_EVAL_NOTE = 0
START_PENDING_EVAL_NOTE = "0"

# tentacles
TENTACLE_DEFAULT_CONFIG = "default_config"
CONFIG_TENTACLES_FILE = "tentacles_config.json"
EVALUATOR_PRIORITY = "priority"
DEFAULT_EVALUATOR_PRIORITY = 0
CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT = "required_candles_count"
NESTED_TENTACLE_CONFIG = "nested_tentacle_configuration"
CONFIG_ACTIVATION_TOPICS = "activation method"
CONFIG_TRIGGER_TIMEFRAMES = "Trigger_timeframes"
CONFIG_EMIT_TRADING_SIGNALS = "emit_trading_signals"
CONFIG_TRADING_SIGNALS_STRATEGY = "trading_strategy"

# terms of service
CONFIG_ACCEPTED_TERMS = "accepted_terms"

# metrics
CONFIG_METRICS = "metrics"
CONFIG_METRICS_BOT_ID = "metrics-bot-id"
TIMER_BEFORE_METRICS_REGISTRATION_SECONDS = 600
TIMER_BETWEEN_METRICS_UPTIME_UPDATE = 3600 * 4
METRICS_URL = os.getenv("METRICS_OCTOBOT_ONLINE_URL", "https://metrics.octobot.online/")
METRICS_ROUTE_GEN_BOT_ID = "gen-bot-id"
METRICS_ROUTE = "metrics"
METRICS_ROUTE_COMMUNITY = f"{METRICS_ROUTE}/community"
METRICS_ROUTE_UPTIME = f"{METRICS_ROUTE}/uptime"
METRICS_ROUTE_REGISTER = f"{METRICS_ROUTE}/register"
COMMUNITY_TOPS_COUNT = 1000

# default values in config files and interfaces
DEFAULT_API_KEY = "your-api-key-here"
DEFAULT_API_SECRET = "your-api-secret-here"
DEFAULT_API_PASSWORD = "your-api-password-here"
DEFAULT_EXCHANGE_TYPE = CONFIG_EXCHANGE_SPOT
DEFAULT_CONFIG_VALUES = {
    DEFAULT_API_KEY,
    DEFAULT_API_SECRET,
    DEFAULT_API_PASSWORD,
    "NOKEY",
    "NO KEY",
    "Empty",
}

# cache
CACHE_FOLDER = "cache"
CACHE_FILE = "cache.json"
CACHE_HASH_SIZE = 15
CACHE_RELATED_DATA_SEPARATOR = "##"
LOCAL_BOT_DATA = "local_bot_data"
DO_NOT_CACHE = "do_not_cache"
DO_NOT_OVERRIDE_CACHE = "do_not_override_cache"
DEFAULT_IGNORED_VALUE = -1
UNPROVIDED_CACHE_IDENTIFIER = "_unprovided"

# Async settings
DEFAULT_FUTURE_TIMEOUT = 120

# Github urls
GITHUB_RAW_CONTENT_URL = "https://raw.githubusercontent.com"
GITHUB_API_CONTENT_URL = "https://api.github.com"
GITHUB_BASE_URL = "https://github.com"
GITHUB_ORGANISATION = "Drakkar-Software"

# External resources
EXTERNAL_RESOURCE_URL = "https://raw.githubusercontent.com/Drakkar-Software/OctoBot/assets/external_resources.json"

# Run databases
DATA_FOLDER = "data"
DB_SEPARATOR = "_"
TINYDB_EXT = ".json"
MAX_BACKTESTING_RUNS = 50000
MAX_OPTIMIZER_RUNS = 50000



# pylint: disable=R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_commons.databases import global_storage
from octobot_commons.databases import database_caches
from octobot_commons.databases import document_database_adaptors
from octobot_commons.databases import bases
from octobot_commons.databases import implementations
from octobot_commons.databases import relational_databases

from octobot_commons.databases import cache_manager
from octobot_commons.databases import databases_util
from octobot_commons.databases import cache_client
from octobot_commons.databases import run_databases

from octobot_commons.databases.global_storage import (
    GlobalSharedMemoryStorage,
)

from octobot_commons.databases.database_caches import (
    GenericDatabaseCache,
    ChronologicalReadDatabaseCache,
)

from octobot_commons.databases.document_database_adaptors import (
    AbstractDocumentDatabaseAdaptor,
    TinyDBAdaptor,
)

from octobot_commons.databases.bases import (
    DocumentDatabase,
    BaseDatabase,
)

from octobot_commons.databases.implementations import (
    DBReader,
    DBWriter,
    DBWriterReader,
    MetaDatabase,
    CacheDatabase,
    CacheTimestampDatabase,
)

from octobot_commons.databases.relational_databases import (
    SQLiteDatabase,
    new_sqlite_database,
)

from octobot_commons.databases.run_databases import (
    RunDatabasesIdentifier,
    RunDatabasesProvider,
    init_bot_storage,
    close_bot_storage,
    AbstractRunDatabasesPruner,
    FileSystemRunDatabasesPruner,
    run_databases_pruner_factory,
)

from octobot_commons.databases.cache_manager import (
    CacheManager,
)

from octobot_commons.databases.databases_util import (
    CacheWrapper,
)

from octobot_commons.databases.cache_client import (
    CacheClient,
)


__all__ = [
    "GlobalSharedMemoryStorage",
    "GenericDatabaseCache",
    "ChronologicalReadDatabaseCache",
    "AbstractDocumentDatabaseAdaptor",
    "TinyDBAdaptor",
    "DocumentDatabase",
    "BaseDatabase",
    "MetaDatabase",
    "DBReader",
    "DBWriter",
    "DBWriterReader",
    "CacheDatabase",
    "CacheTimestampDatabase",
    "SQLiteDatabase",
    "new_sqlite_database",
    "RunDatabasesIdentifier",
    "RunDatabasesProvider",
    "init_bot_storage",
    "close_bot_storage",
    "AbstractRunDatabasesPruner",
    "FileSystemRunDatabasesPruner",
    "run_databases_pruner_factory",
    "CacheManager",
    "CacheWrapper",
    "CacheClient",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.signals.signal_bundle as signal_bundle
import octobot_commons.signals.signal as signal


class SignalBundleBuilder:
    DEFAULT_SIGNAL_CLASS = signal.Signal

    def __init__(self, identifier: str):
        self.signals: list = []
        self.identifier: str = identifier
        self.version: str = None
        self.signal_class = self.__class__.DEFAULT_SIGNAL_CLASS
        self.reset()

    def register_signal(self, topic: str, content: dict, **kwargs):
        """
        Store a signal to be packed on build call
        """
        self.signals.append(self.create_signal(topic, content, **kwargs))

    def create_signal(self, topic: str, content: dict, **kwargs):
        """
        Create a signal from self.signal_class
        """
        return self.signal_class(topic, content, **kwargs)

    def is_empty(self) -> bool:
        """
        Return True when no signal are to be built
        """
        return not self.signals

    def build(self) -> signal_bundle.SignalBundle:
        """
        Create a signal_bundle.SignalBundle from registered signals
        """
        return signal_bundle.SignalBundle(
            self.identifier,
            signals=self.signals,
            version=self.version,
        )

    def reset(self):
        """
        Remove all registered signals
        """
        self.signals = []



# pylint: disable=R0902,C0103
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import os

import octobot_commons.databases.implementations.db_writer_reader as db_writer_reader
import octobot_commons.enums as enums


class MetaDatabase:
    def __init__(self, run_dbs_identifier, with_lock=False, cache_size=None):
        self.run_dbs_identifier = run_dbs_identifier
        self.with_lock = with_lock
        self.cache_size = cache_size
        self.database_adaptor = self.run_dbs_identifier.database_adaptor
        self.run_db: db_writer_reader.DBWriterReader = None
        self.orders_db: db_writer_reader.DBWriterReader = None
        self.trades_db: db_writer_reader.DBWriterReader = None
        self.transactions_db: db_writer_reader.DBWriterReader = None
        self.historical_portfolio_value_db: db_writer_reader.DBWriterReader = None
        self.backtesting_metadata_db: db_writer_reader.DBWriterReader = None
        self.symbol_dbs: dict = {}

    def get_run_db(self):
        """
        :return: the run database. Opens it if not open already
        """
        if self.run_db is None:
            self.run_db = self._get_db(
                self.run_dbs_identifier.get_run_data_db_identifier()
            )
        return self.run_db

    def get_orders_db(self, exchange=None):
        """
        :return: the orders database. Opens it if not open already
        """
        if self.orders_db is None:
            self.orders_db = self._get_db(
                self.run_dbs_identifier.get_orders_db_identifier(
                    exchange or self.run_dbs_identifier.context.exchange_name
                )
            )
        return self.orders_db

    def get_trades_db(self, exchange=None):
        """
        :return: the trades database. Opens it if not open already
        """
        if self.trades_db is None:
            self.trades_db = self._get_db(
                self.run_dbs_identifier.get_trades_db_identifier(
                    exchange or self.run_dbs_identifier.context.exchange_name
                )
            )
        return self.trades_db

    def get_transactions_db(self, exchange=None):
        """
        :return: the transactions database. Opens it if not open already
        """
        if self.transactions_db is None:
            self.transactions_db = self._get_db(
                self.run_dbs_identifier.get_transactions_db_identifier(
                    exchange or self.run_dbs_identifier.context.exchange_name
                )
            )
        return self.transactions_db

    def get_historical_portfolio_value_db(self, exchange, portfolio_type_suffix):
        """
        :return: the historical portfolio database. Opens it if not open already
        """
        if self.historical_portfolio_value_db is None:
            self.historical_portfolio_value_db = self._get_db(
                self.run_dbs_identifier.get_historical_portfolio_value_db_identifier(
                    exchange, portfolio_type_suffix
                )
            )
        return self.historical_portfolio_value_db

    def get_backtesting_metadata_db(self):
        """
        :return: the backtesting metadata database. Opens it if not open already
        """
        if self.backtesting_metadata_db is None:
            self.backtesting_metadata_db = self._get_db(
                self.run_dbs_identifier.get_backtesting_metadata_identifier()
            )
        return self.backtesting_metadata_db

    async def get_backtesting_metadata_from_run(self):
        """
        :return: the backtesting metadata for the associated run_dbs_identifier's backtesting_id
        """
        db = self.get_backtesting_metadata_db()
        return (
            await db.select(
                enums.CacheDatabaseTables.METADATA.value,
                (await db.search()).id == self.run_dbs_identifier.backtesting_id,
            )
        )[0]

    def get_symbol_db(self, exchange, symbol):
        """
        :return: the symbol database. Opens it if not open already
        """
        key = self._get_symbol_db_key(exchange, symbol)
        if key not in self.symbol_dbs:
            self.symbol_dbs[key] = self._get_db(
                self.run_dbs_identifier.get_symbol_db_identifier(exchange, symbol)
            )
        return self.symbol_dbs[key]

    async def get_all_symbol_dbs(self, exchange):
        """
        :return: an iterable over each symbol database for the given exchange
        """
        if self.run_dbs_identifier.database_adaptor.is_file_system_based():
            return [
                self.get_symbol_db(
                    exchange, self.run_dbs_identifier.get_symbol_db_name(db.name)
                )
                for db in os.scandir(
                    self.run_dbs_identifier.get_exchange_based_identifier(exchange)
                )
                if self.run_dbs_identifier.is_symbol_database(db.name)
            ]
        raise NotImplementedError(
            "get_all_symbol_dbs is not implemented for non is_file_system_based databases"
        )

    def all_basic_run_db(self, exchange=None):
        """
        yields the run, orders, trades and transactions databases
        """
        yield self.get_run_db()
        yield self.get_orders_db(exchange)
        yield self.get_trades_db(exchange)
        yield self.get_transactions_db(exchange)

    @staticmethod
    def _get_symbol_db_key(exchange, symbol):
        return f"{exchange}{symbol}"

    def _get_db(self, db_identifier):
        return db_writer_reader.DBWriterReader(
            db_identifier,
            with_lock=self.with_lock,
            cache_size=self.cache_size,
            database_adaptor=self.database_adaptor,
            enable_storage=self.run_dbs_identifier.enable_storage,
        )

    async def close(self):
        """
        Closes all the open databases
        """
        # avoid asyncio.gather here as it is producing unexplained side effects (frozen thread preventing stop)
        for coro in (
            db.close()
            for db in (
                self.run_db,
                self.orders_db,
                self.trades_db,
                self.transactions_db,
                self.historical_portfolio_value_db,
                self.backtesting_metadata_db,
                *self.symbol_dbs.values(),
            )
            if db is not None
        ):
            await coro

    @classmethod
    @contextlib.asynccontextmanager
    async def database(cls, database_manager, with_lock=False, cache_size=None):
        """
        Created a local meta database and closes it upon leaving the context manager
        """
        meta_db = None
        try:
            meta_db = MetaDatabase(
                database_manager, with_lock=with_lock, cache_size=cache_size
            )
            yield meta_db
        finally:
            if meta_db is not None:
                await meta_db.close()



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums
import octobot_commons.signals as signals

from tests.signals import signal_builder_wrapper


def test_register_user(signal_builder_wrapper):
    assert signal_builder_wrapper._users_count == 0
    signal_builder_wrapper.register_user()
    assert signal_builder_wrapper._users_count == 1
    signal_builder_wrapper.register_user()
    assert signal_builder_wrapper._users_count == 2


def test_unregister_user(signal_builder_wrapper):
    assert signal_builder_wrapper._users_count == 0
    signal_builder_wrapper.register_user()
    assert signal_builder_wrapper._users_count == 1
    signal_builder_wrapper.register_user()
    assert signal_builder_wrapper._users_count == 2
    signal_builder_wrapper.unregister_user()
    assert signal_builder_wrapper._users_count == 1
    signal_builder_wrapper.register_user()
    assert signal_builder_wrapper._users_count == 2
    signal_builder_wrapper.unregister_user()
    assert signal_builder_wrapper._users_count == 1
    signal_builder_wrapper.unregister_user()
    assert signal_builder_wrapper._users_count == 0


def test_has_single_user(signal_builder_wrapper):
    assert signal_builder_wrapper.has_single_user() is False
    signal_builder_wrapper._users_count = 1
    assert signal_builder_wrapper.has_single_user() is True
    signal_builder_wrapper._users_count = 10
    assert signal_builder_wrapper.has_single_user() is False
    signal_builder_wrapper._users_count = 1
    assert signal_builder_wrapper.has_single_user() is True
